<?php

namespace Drupal\commerce_opayo_pi\Plugin\Commerce\CheckoutFlow;

use Drupal\commerce_checkout\Plugin\Commerce\CheckoutFlow\CheckoutFlowWithPanesBase;
use Drupal\commerce_checkout\Event\CheckoutEvents;
use Drupal\commerce_order\Event\OrderEvent;
use Drupal\Core\Form\FormStateInterface;


/**
 * @CommerceCheckoutFlow(
 *  id = "opayo_checkout_flow",
 *  label = @Translation("Opayo checkout flow"),
 * )
 */
class OpayoCheckoutFlow extends CheckoutFlowWithPanesBase {

  /**
   * Log a message at the payment gateway's 'extra verbose debug' level
   *
   * @param string|\Stringable $message
   *   Log message
   * @param array $context
   *   Additional info
   */
  public function logVerboseDebug(string|\Stringable $message, array $context = []) {
    if ($this->order != null && $this-> order->get('payment_gateway') != null && $this->order->get('payment_gateway')->entity != null && $this->order->get('payment_gateway')->entity->getPlugin() != null) {
      /** @var \Drupal\commerce_opayo_pi\Plugin\Commerce\PaymentGateway\OpayoPiPaymentGatewayInterface $payment_gateway_plugin */
      $payment_gateway_plugin = $this->order->get('payment_gateway')->entity->getPlugin();
      $checkout_flow = $this->order->get('checkout_flow')->entity;
      $payment_gateway_plugin->logVerboseDebug('OpayoCheckoutFlow(' . $checkout_flow->id() . ':' . $this->pluginId . ')::' . $message, $context);
    }
  }


  /**
   * {@inheritdoc}
   */
  public function getSteps() {
    // Note that previous_label and next_label are not the labels
    // shown on the step itself. Instead, they are the labels shown
    // when going back to the step, or proceeding to the step.
    return [
      'login' => [
        'label' => $this->t('Login'),
        'previous_label' => $this->t('Go back'),
        'has_sidebar' => FALSE,
      ],
      'order_information' => [
        'label' => $this->t('Order information'),
        'has_sidebar' => TRUE,
        'previous_label' => $this->t('Go back'),
      ],
      'review' => [
        'label' => $this->t('Review'),
        'next_label' => $this->t('Continue to review'),
        'previous_label' => $this->t('Go back'),
        'has_sidebar' => TRUE,
      ],
      'payment' => [
        'label' => $this->t('Payment'),
        'next_label' => $this->t('Pay and complete purchase'),
        'has_sidebar' => FALSE,
        'hidden' => TRUE,
      ],
      'threedsecure' => [
        'label' => $this->t('3D Secure'),
        'next_label' => $this->t('Continue'),
        'previous_label' => $this->t('Go back'),
        'has_sidebar' => TRUE,
      ],
      'complete' => [
        'label' => $this->t('Complete'),
        'next_label' => $this->t('Complete checkout'),
        'has_sidebar' => FALSE,
      ],
    ];
  }


  /**
   * {@inheritdoc}
   */
  protected function onStepChange($step_id) {
    $order = $this->getOrder();

    $this->logVerboseDebug('onStepChange step id: ' . $step_id);

    // Lock the order while on the 'payment' checkout step. Unlock elsewhere.
    if ($step_id == 'payment' || $step_id == 'threedsecure') {
      $order->lock();
    }
    elseif ($step_id != 'payment' && $step_id != 'threedsecure') {
      $order->unlock();
    }
    // Place the order.
    if ($step_id == 'complete' && $order->getState()->getId() == 'draft') {
      // Notify other modules.
      $event = new OrderEvent($order);
      $this->eventDispatcher->dispatch($event, CheckoutEvents::COMPLETION);
      $order->getState()->applyTransitionById('place');
    }
  }


  public function onStepChangeCustom($step_id) {
    $this->onStepChange($step_id);
  }

  public function setOrder($order) {
    $this->order = $order;
  }


  /**
   * {@inheritdoc}
   */
  public function getNextStepId($step_id) {
    $next_step = null;
    if ($step_id == 'complete') {
      // There is no step after complete, return 'complete'
      $next_step = $step_id;
    }
    else {
      $next_step = parent::getNextStepId($step_id);
    }

    $this->logVerboseDebug('getNextStepId current: ' . $step_id . ', next: ' . $next_step);
    return $next_step;
  }


  /**
   * {@inheritdoc}
   */
  public function getVisibleSteps() {

    // In the context of a webform order we may need to selective include/exclude the 'payment information'
    // Normally it's excluded but if there has been a failed payment attempt we need to include it

    if (empty($this->visibleSteps)) {
      $steps = $this->getSteps();
      foreach ($steps as $step_id => $step) {
        if (!$this->isStepVisible($step_id)) {
          unset($steps[$step_id]);
        }
      }
      $this->visibleSteps = $steps;
    }

    return $this->visibleSteps;
  }


  /**
   * {@inheritdoc}
   */
  public function redirectToStep($step_id) {
    $this->logVerboseDebug('redirectToStep - step: ' . $step_id);
    parent::redirectToStep($step_id);
  }


  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state, $step_id = NULL) {
    $this->logVerboseDebug('buildForm - step: ' . $step_id);
    $form = parent::buildForm($form, $form_state, $step_id);
    return $form;
  }


}
