<?php

namespace Drupal\commerce_opayo_pi\Plugin\QueueWorker;

use DateTime;
use Drupal\commerce_opayo_pi\OpayoPi;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\Core\Queue\QueueWorkerBase;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Drupal\Core\Logger\LoggerChannelInterface;
use Symfony\Component\Validator\Constraints\Count;

/**
 * Queue job for getting bank settlement and fraud scoring info.
 *
 * Need to add the following to 'settings.php' so that the 'unique queue' gets picked up during cron processing
 * $settings['queue_service_commerce_opayo_pi_daily'] = 'queue_unique.database';
 *
 * @QueueWorker(
 *  id = "commerce_opayo_pi_daily",
 *  title = @Translation("Opayo Pi"),
 *  cron = {"time" = 30}
 * )
 */
class OpayoPiWorkerDaily extends QueueWorkerBase implements ContainerFactoryPluginInterface {

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * Logger
   *
   * @var \Drupal\Core\Logger\LoggerChannelInterface
   */
  protected $logger;

    /**
   * The module's service
   *
   * @var \Drupal\commerce_opayo_pi\OpayoPi
   */
  protected $opayoPiService;

  /**
   * Constructs a new OpayoPiWorkerDaily object.
   *
   * @param array $configuration
   *   A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *   The plugin ID for the plugin instance.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   */
  public function __construct(array $configuration, $plugin_id, $plugin_definition, EntityTypeManagerInterface $entity_type_manager, LoggerChannelInterface $logger, OpayoPi $opayoPi) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);

    $this->entityTypeManager = $entity_type_manager;
    $this->logger = $logger;
    $this->opayoPiService = $opayoPi;
  }


  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('entity_type.manager'),
      $container->get('commerce_opayo_pi.logger.channel.commerce_opayo_pi'),
      $container->get('commerce_opayo_pi.opayo_pi'),
    );
  }

  public function processItem($data) {
    if (property_exists($data, 'opayoBatchDateToProcess')) {
      // Fetch transaction batch info
      $this->processOpayoBatchDate($data->opayoBatchDateToProcess);
    }
  }

  /**
   * Retrieve bank settlement batch information for a given date and update the 'batch ID' field on OpayoTransaction entities
   *
   * @param string $opayoBatchDateToProcess
   *   The date for which to get batch IDs
   *
   */
  protected function processOpayoBatchDate(string $opayoBatchDateToProcess) {

    $opayo_config = $this->opayoPiService->getSettings();

    $url = $opayo_config == null ? null : $opayo_config['liveaccess'];    // Batch IDs are only created for 'live' payments, not test
    $account_user = $opayo_config == null ? null : $opayo_config['account_user'];
    $account_password = $opayo_config == null ? null : $opayo_config['account_password'];

    if (isset($account_user) && isset($account_password) && isset($url)) {

      // First, get the 'batch ID' associated with the date by using the 'getBatchList' command
      $batch_ids = array();
      $batch_completed = array();

      // Use Opayo's 'Reporting & Admin API', see https://developer.elavon.com/products/opayo-reporting-api/v1

      $parse_result = date_parse($opayoBatchDateToProcess);
      $start_date = DateTime::createFromFormat('Y-m-d H:i:s', strval($parse_result['year']) . '-' . strval($parse_result['month']) . '-' . strval($parse_result['day']) . '00:00:00');
      $end_date = clone $start_date;
      $end_date->modify('+ 1 day');
      $start_date_xml = '<startdate>' . $start_date->format('d/m/Y H:i:s') . '</startdate>';
      $end_date_xml = '<enddate>' . $end_date->format('d/m/Y H:i:s') . '</enddate>';
      $request_contents = $this->requestXML($opayo_config['vendor'], $account_user, $account_password, 'getBatchList', $start_date_xml . $end_date_xml);

      $curl = curl_init();
      curl_setopt_array($curl, array(
        CURLOPT_URL => $url,
        CURLOPT_RETURNTRANSFER => true,
        CURLOPT_CUSTOMREQUEST => "POST",
        CURLOPT_POSTFIELDS => http_build_query(["XML" => $request_contents]),
        CURLOPT_HTTPHEADER => array(
          "Cache-Control: no-cache",
          "Content-Type: application/x-www-form-urlencoded"
        ),
      ));

      $this->logger->debug('OpayoPiWorker::processOpayoBatchDate: about to request getBatchList for date: {batchDate} ', ['batchDate' => $start_date->format('d/m/Y')]);
      $raw_response = curl_exec($curl);
      $http_code = curl_getinfo($curl, CURLINFO_RESPONSE_CODE);
      $err = curl_error($curl);

      if (!empty($err)) {
        $this->logger->warning('OpayoPiWorker::processOpayoBatchDate: Unsuccessful getBatchList request date: {batchDate}, error: {error} ', ['batchDate' => $start_date->format('d/m/Y'), 'error' => $err]);
      } else {
        $this->logger->debug('OpayoPiWorker::processOpayoBatchDate: Successful getBatchList request date: {batchDate} ', ['batchDate' => $start_date->format('d/m/Y')]);
      }

      $contents = json_decode(json_encode(simplexml_load_string($raw_response)), true);

      curl_close($curl);

      if (array_key_exists('errorcode', $contents)) {
        $error_code = $contents['errorcode'];

        if ($error_code == '0000') {
          // success
          $transactions = $contents['batches'];
          foreach($transactions as $transaction) {
            $batch_id = $transaction['batchid'];
            $auth_processor = $transaction['authprocessor'];
            $batch_completed = $transaction['completed'];
            // Add to the list of batch IDs for which we need to get details
            $batch_ids[$batch_id] = $auth_processor;
            $batch_completed[$batch_id] = DateTime::createFromFormat('d/m/Y H:i:s', $batch_completed);
            $this->logger->info('OpayoPiWorker::processOpayoBatchDate: Batch ID is {batchId} for date: {batchDate} ', ['batchDate' => $start_date->format('d/m/Y'), 'batchId' => $batch_id]);
          }
        }
        else {
          $this->logger->warning('OpayoPiWorker::processOpayoBatchDate: getBatchList returns error for request date: {batchDate}, code: {error} ', ['batchDate' => $start_date->format('d/m/Y'), 'error' => $error_code]);
        }
      }
      else {
        $this->logger->warning('OpayoPiWorker::processOpayoBatchDate: getBatchList unexpected response for request date: {batchDate}, payload: {payload} ', ['batchDate' => $start_date->format('d/m/Y'), 'payload' => $raw_response]);
      }

      if (count($batch_ids) == 0) {
        $this->logger->info('OpayoPiWorker::processOpayoBatchDate: getBatchList returns no batch request date: {batchDate}', ['batchDate' => $start_date->format('d/m/Y')]);
      }
      else if (count($batch_ids) > 0) {
        // Transactions matching the batch
        $opayo_transactions = array();

        foreach(array_keys($batch_ids) as $batch_id) {

          $auth_processor = $batch_ids[$batch_id];
          $batch_id_xml = '<batchid>' . $batch_id . '</batchid>';
          $auth_processor_xml = '<authprocessor>' . $auth_processor . '</authprocessor>';
          $request_contents = $this->requestXML($opayo_config['vendor'], $account_user, $account_password, 'getBatchDetail', $batch_id_xml . $auth_processor_xml);

          $curl = curl_init();
          curl_setopt_array($curl, array(
            CURLOPT_URL => $url,
            CURLOPT_RETURNTRANSFER => true,
            CURLOPT_CUSTOMREQUEST => "POST",
            CURLOPT_POSTFIELDS => http_build_query(["XML" => $request_contents]),
            CURLOPT_HTTPHEADER => array(
              "Cache-Control: no-cache",
              "Content-Type: application/x-www-form-urlencoded"
            ),
          ));

          $this->logger->debug('OpayoPiWorker::processOpayoBatchDate: about to request getBatchDetail for ID: {batchId} ', ['batchId' => $batch_id]);
          $raw_response = curl_exec($curl);
          $http_code = curl_getinfo($curl, CURLINFO_RESPONSE_CODE);
          $err = curl_error($curl);

          if (!empty($err)) {
            $this->logger->warning('OpayoPiWorker::processOpayoBatchDate: Unsuccessful getBatchDetail ID: {batchId} , error: {error} ', ['batchId' => $batch_id, 'error' => $err]);
          } else {
            $this->logger->debug('OpayoPiWorker::processOpayoBatchDate: Successful getBatchDetail ID: {batchId} ', ['batchId' => $batch_id]);
          }

          $response_array = json_decode(json_encode(simplexml_load_string($raw_response)), true);

          curl_close($curl);

          if (array_key_exists('vspaccess', $response_array)) {
            $contents = $response_array['vspaccess'];
            $error_code = $contents['errorcode'];

            if ($error_code == '0000') {
              // success
              $transactions = $contents['transactions'];
              foreach ($transactions as $transaction) {
                $opayo_transaction_id = $transaction['vpstxid'];
                $opayo_transaction_time = $transactions['started'];
                $opayo_transaction_amount = $transactions['amount'];

                // Collect additional 3rd man and fraud scoring info
                $additional_info = array();
                if (array_key_exists('t3mscore', $transaction))
                  $additional_info['3rd man score'] = $transaction['t3mscore'];
                if (array_key_exists('t3maction', $transaction))
                  $additional_info['3rd man score'] = $transaction['t3maction'];
                if (array_key_exists('fraudscreenrecommendation', $transaction))
                  $additional_info['3rd man score'] = $transaction['fraudscreenrecommendation'];
                if (array_key_exists('fraudcode', $transaction))
                  $additional_info['3rd man score'] = $transaction['fraudcode'];
                if (array_key_exists('fraudcodedetail', $transaction))
                  $additional_info['3rd man score'] = $transaction['fraudcodedetail'];
                // Save additional info with the transaction ID
                $opayo_transactions[$opayo_transaction_id] = $additional_info;

                $this->logger->info('OpayoPiWorker::processOpayoBatchDate: Transaction {transactionId} dated {transTime} for amount {amount} part of batch: {batchId} ', ['batchId' => $batch_id, 'batchId' => $batch_id, 'transactionId' => $opayo_transaction_id, 'transTime' => $opayo_transaction_time, 'amount' => $opayo_transaction_amount]);
              }
            } else {
              $this->logger->warning('OpayoPiWorker::processOpayoBatchDate: getBatchDetail returns error for ID: {batchId}, code: {error} ', ['batchId' => $batch_id, 'error' => $error_code]);
            }
          } else {
            $this->logger->warning('OpayoPiWorker::processOpayoBatchDate: getBatchDetail unexpected response for ID: {batchId}, payload: {payload} ', ['batchId' => $batch_id, 'payload' => $raw_response]);
          }
        }

        foreach(array_keys($opayo_transactions) as $opayo_transaction) {
          // Match
          $entity = $this->entityTypeManager->getStorage('opayo_transaction')->load($opayo_transactions);
          if ($entity != null) {
            /** @var \Drupal\commerce_opayo_pi\Entity\OpayoTransaction $opayo_transaction */
            $opayo_transaction = $entity;

            // Set the batch ID, batch completed timestamp and optionally info about fraud scoring
            $opayo_transaction->set('batch_id', $batch_id);
            $opayo_transaction->set('batch_completed', $batch_completed[$batch_id]);
            if (count($opayo_transactions[$opayo_transaction_id]) > 0) {
              $opayo_transaction->set('fraud_score_info', implode(', ', $opayo_transactions[$opayo_transaction_id]));
            }
            $opayo_transaction->save();

            $this->logger->info('OpayoPiWorker::processOpayoBatchDate: Transaction entity {transactionId} set batch ID TO: {batchId} ', ['batchId' => $batch_id, 'batchId' => $batch_id, 'transactionId' => $opayo_transaction_id]);
          }
        }
      }
    }
  }


  /**
   * Creates an XML request message
   *
   * @param array $vendor
   *   Opayo 'vendor'
   * @param string $user
   *   Opayo account username
   * @param string $password
   *   Opayo account password
   * @param string $command
   *   Command to submit - see API
   * @param string $other
   *   Additional XML fields to support the command
   *
   * @return string
   *   request message in XML format
   */
  protected function requestXML($vendor, $user, $password, $command, $other = null) {
    $xml_string_contents = '<command>' . $command . '</command>' .
                   '<vendor>' . $vendor . '</vendor>' .
                   '<user>' . $user . '</user>' .
                   (isset($other) ? $other : '');

    // MD5 Hash the contents with the password included
    $xml_string_contents_hash = md5($xml_string_contents . '<password>'. $password . '</password>');

    // Return contents and the MD5 hash signature
    return '<vspaccess>' . $xml_string_contents . '<signature>' . $xml_string_contents_hash . '</signature></vspaccess>';
  }
}
